
/*

This clock uses an animation that scrolls up digits that change.
Displayed below is the animation schema for all the digits of the clock face.

  animation frame counter (counting up)
  0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 2 2 2 2 2 2 2 2 2 2
  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9

H	- - - - - - - - - - - - - - - - - - 1 2 3 4 5 6 7 8 9 - - -   [18,26]
H	- - - - - - - - - - - - 1 2 3 4 5 6 7 8 9 - - - - - - - - -   [12,20]
M	- - - - - - 1 2 3 4 5 6 7 8 9 - - - - - - - - - - - - - - -   [6,14]
M	1 2 3 4 5 6 7 8 9 - - - - - - - - - - - - - - - - - - - - -   [0,8]

*/

#include <PimientoClockLib.h>
#include <Streaming.h>

const uint8_t StateClock1 = 0;
const uint8_t StateClock2 = 1;
const uint8_t StateClock3 = 2;
const uint8_t StateSetHours = 3;
const uint8_t StateSetMinutes = 4;
const uint8_t StateResetSeconds = 5;

uint8_t currentState = StateClock2;
bool stateInitNeeded = true;
uint32_t stateUpdateCounter = 0;

extern void updateState(uint8_t event = EventSystemUpdate);

uint32_t currentTime = 0;
uint32_t newTime = 0;

// ----

void setup() {
  pimiento.init(updateState);
}

void loop() {
  pimiento.loop();
}

// ----

void setState(uint8_t newState) {
  stateInitNeeded = true;
  stateUpdateCounter = 0;
  currentState = newState;
}

void updateState(uint8_t event) {

  if (event != EventSystemUpdate) {
    Serial << "updateState - currentState: " << currentState << " - event: " << event << endl;
  }

  switch (currentState) {
    case StateClock1:
      if (event == EventSystemUpdate) PimientoScheduler::everyMillis(250, stateUpdateCounter, loopDrawClock1);
      else if (event == EventCenterButtonReleased) setState(StateClock2);
      else if (event == EventLeftButtonReleased) incBrightness(-1);
      else if (event == EventRightButtonReleased) incBrightness(1);
      break;
    case StateSetHours:
      if (event == EventSystemUpdate) PimientoScheduler::everyMillis(250, stateUpdateCounter, loopDrawClock1);
      else if (event == EventLeftButtonReleased) incHour(-1);
      else if (event == EventRightButtonReleased) incHour(1);
      else if (event == EventCenterButtonReleased) setState(StateSetMinutes);
      break;
    case StateSetMinutes:
      if (event == EventSystemUpdate) PimientoScheduler::everyMillis(250, stateUpdateCounter, loopDrawClock1);
      else if (event == EventLeftButtonReleased) incMinute(-1);
      else if (event == EventRightButtonReleased) incMinute(1);
      else if (event == EventCenterButtonReleased) setState(StateResetSeconds);
      break;
    case StateResetSeconds:
      if (event == EventSystemUpdate) PimientoScheduler::everyMillis(250, stateUpdateCounter, loopDrawClock1);
      else if (event == EventLeftButtonReleased) resetSeconds();
      else if (event == EventRightButtonReleased) resetSeconds();
      else if (event == EventCenterButtonReleased) setState(StateClock1);
      break;
    case StateClock2:
      if (event == EventSystemUpdate) PimientoScheduler::everyMillis(100, stateUpdateCounter, loopDrawClock2);
      else if (event == EventCenterButtonReleased) setState(StateClock3);
      else if (event == EventLeftButtonReleased) incBrightness(-1);
      else if (event == EventRightButtonReleased) incBrightness(1);
      break;
    case StateClock3:
      if (event == EventSystemUpdate) PimientoScheduler::everyMillis(250, stateUpdateCounter, loopDrawClock3);
      else if (event == EventCenterButtonReleased) setState(StateSetHours);
      break;
    default:
      Serial << "unsupported - state: " << currentState << ", event: " << event << endl;
  }
}

void incBrightness(int8_t increment) {

  int8_t brightness = (display.getBrightness() + increment);
  if (brightness == -1) {
    brightness = MAX_BRIGHTNESS;
  }
  if (brightness > MAX_BRIGHTNESS) {
    brightness = 0;
  }
  display.setBrightness(brightness);
}

void incHour(int8_t increment) {
  rtc.refresh();
  int8_t newHour = (rtc.hour() + increment) % 24;
  rtc.set(rtc.second(), rtc.minute(), newHour, rtc.dayOfWeek(), rtc.day(), rtc.month(), rtc.year());
}

void incMinute(int8_t increment) {
  rtc.refresh();
  int8_t newMinute = (rtc.minute() + increment) % 60;
  rtc.set(rtc.second(), newMinute, rtc.hour(), rtc.dayOfWeek(), rtc.day(), rtc.month(), rtc.year());
}

void resetSeconds() {
  rtc.refresh();
  rtc.set(0, rtc.minute(), rtc.hour(), rtc.dayOfWeek(), rtc.day(), rtc.month(), rtc.year());
}

void loopDrawClock1() {

  rtc.refresh();

  display.clearAll();
  display.setPage(0);
  display.setCursor(1, 1);

  uint8_t hoursColor = (currentState == StateSetHours) ? MAIN_COLOR + BLINK : MAIN_COLOR;
  display.printInt(rtc.hour(), 2, hoursColor);

  display.moveCursor(1, 0);

  uint8_t minutesColor = (currentState == StateSetMinutes) ? MAIN_COLOR + BLINK : MAIN_COLOR;
  display.printInt(rtc.minute(), 2, minutesColor);

  display.moveCursor(1, 0);

  uint8_t secondsColor = (currentState == StateResetSeconds) ? MAIN_COLOR + BLINK : MAIN_COLOR;
  display.printInt(rtc.second(), 2, secondsColor);
}

bool animationActive = false;
uint32_t animationDelayCounter = 0;
uint16_t animationRepeatCounter = 0;
uint8_t animationFrame = 0;

bool minutesOnesAnimationNeeded = false;
bool minutesTensAnimationNeeded = false;
bool hoursOnesAnimationNeeded = false;
bool hoursTensAnimationNeeded = false;

void loopDrawClock2() {

  // update time
  rtc.refresh();
  newTime = 100 * rtc.hour() + rtc.minute();

  // first time, initialize display
  if (stateInitNeeded) {
    Serial << "stateInitNeeded" << endl;

    currentTime = newTime;
    stateInitNeeded = false;

    // display new time on display page #0 
    displayTimeOnPageZero();

    animationActive = false;
    animationFrame = 0;

    return;
  }

  // time string has been updated
  // if (newTime != currentTime || rtc.second() % 5 == 0) {
  if (newTime != currentTime) {

    Serial << "newTime: " << newTime << endl;

    drawTimeOnPage(1);

    setupScrollAnimation();

    currentTime = newTime;

    return;
  }

  // perform another animation frame
  if (animationActive && animationRepeatCounter > 0) {
    Serial << "animationRepeatCounter: " << animationRepeatCounter << endl;
    PimientoScheduler::repeat(0, animationDelayCounter, animationRepeatCounter, scrollPage);

    return;
  }

  // animation is fully finished
  if (animationActive && animationRepeatCounter == 0) {

    animationActive = false;
    minutesOnesAnimationNeeded = false;
    minutesTensAnimationNeeded = false;
    hoursOnesAnimationNeeded = false;
    hoursTensAnimationNeeded = false;

    displayTimeOnPageZero();

    return;
  }
}

void setupScrollAnimation() {
  animationActive = true;
  animationRepeatCounter = 0;
  animationFrame = 0;

  uint8_t currentTimeMinutesOnes = currentTime % 10;
  uint8_t currentTimeMinutesTens = (currentTime / 10) % 10;
  uint8_t currentTimeHoursOnes = (currentTime / 100) % 10;
  uint8_t currentTimeHoursTens = (currentTime / 1000) % 10;

  // Serial << "currentTimeMinutesOnes: " << currentTimeMinutesOnes << endl;
  // Serial << "currentTimeMinutesTens: " << currentTimeMinutesTens << endl;
  // Serial << "currentTimeHoursOnes: " << currentTimeHoursOnes << endl;
  // Serial << "currentTimeHoursTens: " << currentTimeHoursTens << endl;

  uint8_t newTimeMinutesOnes = newTime % 10;
  uint8_t newTimeMinutesTens = (newTime / 10) % 10;
  uint8_t newTimeHoursOnes = (newTime / 100) % 10;
  uint8_t newTimeHoursTens = (newTime / 1000) % 10;

  // Serial << "newTimeMinutesOnes: " << newTimeMinutesOnes << endl;
  // Serial << "newTimeMinutesTens: " << newTimeMinutesTens << endl;
  // Serial << "newTimeHoursOnes: " << newTimeHoursOnes << endl;
  // Serial << "newTimeHoursTens: " << newTimeHoursTens << endl;

  minutesOnesAnimationNeeded = newTimeMinutesOnes != currentTimeMinutesOnes;
  minutesTensAnimationNeeded = newTimeMinutesTens != currentTimeMinutesTens;
  hoursOnesAnimationNeeded = newTimeHoursOnes != currentTimeHoursOnes;;
  hoursTensAnimationNeeded = newTimeHoursTens != currentTimeHoursTens;

  // Serial << "minutesOnesAnimationNeeded: " << minutesOnesAnimationNeeded << endl;
  // Serial << "minutesTensAnimationNeeded: " << minutesTensAnimationNeeded << endl;
  // Serial << "hoursOnesAnimationNeeded: " << hoursOnesAnimationNeeded << endl;
  // Serial << "hoursTensAnimationNeeded: " << hoursTensAnimationNeeded << endl;

  if (minutesOnesAnimationNeeded) {
    animationRepeatCounter += 9;
  }

  if (minutesTensAnimationNeeded) {
    animationRepeatCounter += 6;
  }

  if (hoursOnesAnimationNeeded) {
    animationRepeatCounter += 6;
  }

  if (hoursTensAnimationNeeded) {
    animationRepeatCounter += 6;
  }
}

void scrollPage() {

  if (hoursTensAnimationNeeded && animationFrame >= 18 && animationFrame <= 26) {
    moveBlockUp(0, 8, 0, 17);
  }

  if (hoursOnesAnimationNeeded && animationFrame >= 12 && animationFrame <= 20) {
    moveBlockUp(8, 8, 0, 17);
  }

  if (minutesTensAnimationNeeded && animationFrame >= 6 && animationFrame <= 14) {
    moveBlockUp(16, 8, 0, 17);
  }

  if (minutesOnesAnimationNeeded && animationFrame >= 0 && animationFrame <= 8) {
    moveBlockUp(24, 8, 0, 17);
  }

  animationFrame++;
}

void moveBlockUp(uint8_t startX, uint8_t width, uint8_t startY, uint8_t height) {
  for (uint8_t y = startY; y < (startY + height - 1); y++) {
    display.copyRowContents(startX, width, y + 1, y);
  }
}

void drawTimeOnPage(uint8_t pageNumber) {
    uint8_t savedPage = display.getPage();
    uint8_t savedFont = display.getFont();

    if (pageNumber == 0) {
      display.clearFrameBufferRows(0, 8);
  }

    if (pageNumber == 1) {
      display.clearFrameBufferRows(8, 9);
    }

    display.setFont(2);
    display.setCursor(0, 9 * pageNumber);
    display.printInt(rtc.hour(), 2, MAIN_COLOR);
    display.printInt(rtc.minute(), 2, MAIN_COLOR);
    
    display.setPage(savedPage);
    display.setFont(savedFont);    
}

void displayTimeOnPageZero() {
  drawTimeOnPage(0);
  display.setStartY(0);
}

void loopDrawClock3() {

  rtc.refresh();

  display.clearAll();
  display.setPage(0);
  display.setFont(3);
  display.setCursor(8, 1);
  display.printInt(rtc.hour(), 2, MAIN_COLOR);
  display.moveCursor(1, 0);
  display.printInt(rtc.minute(), 2, MAIN_COLOR);
  display.setFont(0);
}
